package org.cte.ABCD.grammar.tests;

import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.lang.reflect.Method;

import org.antlr.v4.runtime.ANTLRFileStream;
import org.antlr.v4.runtime.ANTLRInputStream;
import org.antlr.v4.runtime.BaseErrorListener;
import org.antlr.v4.runtime.CharStream;
import org.antlr.v4.runtime.CommonTokenStream;
import org.antlr.v4.runtime.Parser;
import org.antlr.v4.runtime.RecognitionException;
import org.antlr.v4.runtime.Recognizer;
import org.antlr.v4.runtime.tree.ParseTree;
import org.cte.ABCD.grammar.ABCDLexer;
import org.cte.ABCD.grammar.ABCDParser;

import org.junit.Test;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.assertEquals;

public class ABCDGrammarTest {
	private String currentRule;
	private Parser mParser;
	
	@Test
	public void testSystemName() {
		setCurrentRule("systemName");
		assertParse("system tot;", "(systemName system tot ;)");
		assertParse("system exemple_stack;", "(systemName system exemple_stack ;)");
	}
	
	@Test
	public void testPrimitiveTypeDecl() {
		setCurrentRule("typeDecl");
		assertParse("type toto is bool", "(typeDecl type toto is (type (primitiveType bool)))");
		assertParse("type toto is nat", "(typeDecl type toto is (type (primitiveType nat)))");
		assertParse("type toto is int", "(typeDecl type toto is (type (primitiveType int)))");
		assertParse("type toto is titi", "(typeDecl type toto is (type (aliasType titi)))");
		assertParse("type toto is none", "(typeDecl type toto is (type (primitiveType none)))");
		
		denyParse("type mix is 5");
		denyParse("type 78 is int");
	}
	
	@Test
	public void testSpecialType() {
		setCurrentRule("type");
		assertParse("timer", "(type (primitiveType timer))");
		assertParse("semaphore", "(type (primitiveType semaphore))");
		assertParse("mutex", "(type (primitiveType mutex))");
	}
	
	@Test
	public void testIntervalType() {
		setCurrentRule("intervalType");
		assertParse("5 .. 10", "(intervalType (expression (literal (integerLiteral 5))) .. (expression (literal (integerLiteral 10))))");
	}
	
	@Test
	public void testUnionType() {
		setCurrentRule("unionType");
		assertParse("union a of int | b, c, d of nat end", "(unionType union " +
																	"(unionFields (unionFieldGroup " +
																		"(identifierList (annotatedIdentifier a)) of (type (primitiveType int))) | " +
																		"(unionFieldGroup (identifierList (annotatedIdentifier b) , (annotatedIdentifier c) , (annotatedIdentifier d)) of (type (primitiveType nat)))) " +
															"end)");
		assertParse("union a of int | b of char | c of nat end union");
	}
	
	@Test
	public void testRecordDeclaration() {
		setCurrentRule("recordType");
		assertParse("record a,b,c : int, d : nat end");
		assertParse("record x : real, y : speci, z : toto end record");
	}
	
	@Test
	public void testArrayDeclaration() {
		setCurrentRule("arrayType");
		assertParse("array 5 of nat", "(arrayType array (expression (literal (integerLiteral 5))) of (type (primitiveType nat)))");
		assertParse("array 6 of array 10 of nat");
	}
	
	@Test
	public void testSynchroPolicy() {
		setCurrentRule("synchronizationPolicy");
		assertParse("sync", "(synchronizationPolicy sync)");
		denyParse("syncnoblock"); //look at this later with a space maybe
		denyParse("syncblock");
		
		assertParse("async block", "(synchronizationPolicy async block)");
		assertParse("async noblock", "(synchronizationPolicy async noblock)");
		assertParse("async block buffersize : 10", "(synchronizationPolicy async block (bufferSize buffersize : (expression (literal (integerLiteral 10)))))");
		assertParse("eventpool buffersize : 10", "(synchronizationPolicy eventpool (bufferSize buffersize : (expression (literal (integerLiteral 10)))))");
		denyParse("async");
		denyParse("async tto");
	}
	
	@Test
	public void testChannelTypeDecl() {
		setCurrentRule("channelTypeDecl");
		assertParse("channeltype t_SyncPort is sync of none", "(channelTypeDecl channeltype t_SyncPort is (synchronizationPolicy sync) of (type (primitiveType none)))");
		assertParse("channeltype t_Asyn is async block buffersize: SIZE_FIFO of int", "(channelTypeDecl channeltype t_Asyn is (synchronizationPolicy async block (bufferSize buffersize : (expression (reference SIZE_FIFO)))) of (type (primitiveType int)))");
		assertParse("channeltype t_As is async noblock of nat");
		denyParse("channeltype tC is sync block of int");
		denyParse("channeltype tS is sync buffersize : 10 of toto");
		denyParse("channeltype Ta is async of int");
		denyParse("channeltype Tb is async buffersize : 4 of int");
	}
	
	@Test
	public void testChannelDecl() {
		setCurrentRule("channelDecl");
		assertParse("channel fifo_wr: tAsty");
		assertParse("channel f : t", "(channelDecl channel (typedIdentifierGroup (identifierList (annotatedIdentifier f)) : (type (aliasType t))))");
		assertParse("channel f, g : t", "(channelDecl channel (typedIdentifierGroup (identifierList (annotatedIdentifier f) , (annotatedIdentifier g)) : (type (aliasType t))))");
	}
	
	@Test 
	public void testBoolConstant() {
		setCurrentRule("constantDecl");
		assertParse("const ACK : bool is true", "(constantDecl const (typedIdentifierGroup (identifierList (annotatedIdentifier ACK)) : (type (primitiveType bool))) is (expression (literal (booleanLiteral true))))");
		assertParse("const NACK : bool is false", "(constantDecl const (typedIdentifierGroup (identifierList (annotatedIdentifier NACK)) : (type (primitiveType bool))) is (expression (literal (booleanLiteral false))))");
		//list of constants initialized with the same value
		assertParse("const a1, a2, a3 : bool is true", "(constantDecl const (typedIdentifierGroup (identifierList (annotatedIdentifier a1) , (annotatedIdentifier a2) , (annotatedIdentifier a3)) : (type (primitiveType bool))) is (expression (literal (booleanLiteral true))))");
	}
	
	@Test
	public void testVariableDecl() {
		setCurrentRule("variableDecl");
		assertParse("var a, b, c : int", "(variableDecl var (variableList (initializedIdentifierGroup (typedIdentifierGroup (identifierList (annotatedIdentifier a) , (annotatedIdentifier b) , (annotatedIdentifier c)) : (type (primitiveType int))))))");
		assertParse("var a, b : int; x : nat; z,w : toto");
		assertParse("var a, b : int := 4; x : nat := 1");
	}
	
	@Test
	public void testPorts() {
		setCurrentRule("ports");
		assertParse("port p : in none", "(ports port (portList (identifierList (annotatedIdentifier p)) : (direction in) (type (primitiveType none))))");
	}
	
	@Test
	public void testArgumentMap() {
		setCurrentRule("argumentMap");
		assertParse("(a, b, c)", "(argumentMap ( (formalActualMap (formalActualAssociation (expression (reference a))) , (formalActualAssociation (expression (reference b))) , (formalActualAssociation (expression (reference c)))) ))");
		assertParse("(4+2)", "(argumentMap ( (formalActualMap (formalActualAssociation (expression (expression (literal (integerLiteral 4))) + (expression (literal (integerLiteral 2)))))) ))");
	}
	
	@Test
	public void testPortList() {
		setCurrentRule("portMap");
		assertParse("[a, b]", "(portMap [ (formalActualMap (formalActualAssociation (expression (reference a))) , (formalActualAssociation (expression (reference b)))) ])");
		assertParse("[a[2]]");
		assertParse("[4]"); //the following lines pass during parsing however they should not be accepted later
		assertParse("[a+2]");
		denyParse("a+b");
	}
	
	@Test
	public void testInstatiation() {
		setCurrentRule("instantiation");
		assertParse("proc(2)[a]", "(instantiation proc (argumentMap ( (formalActualMap (formalActualAssociation (expression (literal (integerLiteral 2))))) )) (outputs (portMap [ (formalActualMap (formalActualAssociation (expression (reference a)))) ])))");
		assertParse("p[b]", "(instantiation p (outputs (portMap [ (formalActualMap (formalActualAssociation (expression (reference b)))) ])))");
		assertParse("p(2)", "(instantiation p (argumentMap ( (formalActualMap (formalActualAssociation (expression (literal (integerLiteral 2))))) )))");
		assertParse("[i]p[o]", "(instantiation (inputs (portMap [ (formalActualMap (formalActualAssociation (expression (reference i)))) ])) p (outputs (portMap [ (formalActualMap (formalActualAssociation (expression (reference o)))) ])))");
		assertParse("p", "(instantiation p)");
	}
	
	@Test
	public void testComposition() {
		setCurrentRule("compositionDecl");
		assertParse("par p || v[b] || d(2) end", "(compositionDecl par (instanceList (instantiation p) || (instantiation v (outputs (portMap [ (formalActualMap (formalActualAssociation (expression (reference b)))) ]))) || (instantiation d (argumentMap ( (formalActualMap (formalActualAssociation (expression (literal (integerLiteral 2))))) )))) end)");
		assertParse("par x end");
		assertParse("par a || b end par");
	}
	
	@Test
	public void testInitDecl() {
		setCurrentRule("initDecl");
		assertParse("init a := 2");
		assertParse("init b := 2; c := 5+2;");
	}
	
	@Test
	public void testIntConstant() {
		setCurrentRule ("constantDecl");
		assertParse("const anInt : int is 100", "(constantDecl const (typedIdentifierGroup (identifierList (annotatedIdentifier anInt)) : (type (primitiveType int))) is (expression (literal (integerLiteral 100))))");
		assertParse("const anotherInt : ins is 100 + 5", "(constantDecl const (typedIdentifierGroup (identifierList (annotatedIdentifier anotherInt)) : (type (aliasType ins))) is (expression (expression (literal (integerLiteral 100))) + (expression (literal (integerLiteral 5)))))");
	}
	
	@Test
	public void testBooleanLiteral() {
		setCurrentRule("booleanLiteral");
		assertParse("true", "(booleanLiteral true)");
		assertParse("false", "(booleanLiteral false)");
		denyParse("fase");
		denyParse("trueb");
	}
	
	@Test
	public void testIntLiteral() {
		setCurrentRule("integerLiteral");
		assertParse("100", "(integerLiteral 100)");
		assertParse("1234", "(integerLiteral 1234)");
		denyParse("true");
	}
	
	@Test
	public void testArrayLiteral() {
		setCurrentRule("arrayLiteral");
		assertParse("[1, 2, 3]", "(arrayLiteral [ (expressionList (expression (literal (integerLiteral 1))) , (expression (literal (integerLiteral 2))) , (expression (literal (integerLiteral 3)))) ])");
		assertParse("[[1, 2, 3], [4, 5, 6], [7, 8, 9]]");
		assertParse("[a+3, b%2, c*d]");
	}
	
	@Test
	public void testRecordLiteral() {
		setCurrentRule("recordLiteral");
		assertParse("{a=1,b=2}", "(recordLiteral { (fieldList (field a = (expression (literal (integerLiteral 1)))) , (field b = (expression (literal (integerLiteral 2))))) })");
		assertParse("{a=1, b={x=1, y=23}, c=23}");
		assertParse("{a=1+3, b=54*tod}");
	}
	
	@Test
	public void testArrayMixRecord() {
		setCurrentRule("literal");
		assertParse("[{x=1, y=2}, {x=6, y=3}]");
		assertParse("{a=[1, 2, 3], b=[{x=2, t=[5, 4]}]}");
	}
	
	@Test
	public void testLiteral() {
		setCurrentRule("literal");
		assertParse("true", "(literal (booleanLiteral true))");
		assertParse("false", "(literal (booleanLiteral false))");
		assertParse("100", "(literal (integerLiteral 100))");
		assertParse("1234", "(literal (integerLiteral 1234))");
		denyParse(".004");
	}
	
	@Test
	public void testUnaryExp() {
		setCurrentRule("field"); //we use field rule since the expression production has one argument
		assertParse("a = not b", "(field a = (expression not (expression (reference b))))");
		assertParse("a = + b", "(field a = (expression + (expression (reference b))))");
		assertParse("a = -b", "(field a = (expression - (expression (reference b))))");
	}
	
	@Test
	public void testBinaryExp() {
		setCurrentRule("field");
		assertParse("a = a*b", "(field a = (expression (expression (reference a)) * (expression (reference b))))");
		assertParse("a = a/b", "(field a = (expression (expression (reference a)) / (expression (reference b))))");
		assertParse("a = a%b", "(field a = (expression (expression (reference a)) % (expression (reference b))))");
		
		assertParse("a = a+b", "(field a = (expression (expression (reference a)) + (expression (reference b))))");
		assertParse("a = a-b", "(field a = (expression (expression (reference a)) - (expression (reference b))))");
		
		assertParse("a = a<b", "(field a = (expression (expression (reference a)) < (expression (reference b))))");
		assertParse("a = a<=b", "(field a = (expression (expression (reference a)) <= (expression (reference b))))");
		assertParse("a = a>b", "(field a = (expression (expression (reference a)) > (expression (reference b))))");
		assertParse("a = a>=b", "(field a = (expression (expression (reference a)) >= (expression (reference b))))");
		
		assertParse("a = a=b", "(field a = (expression (expression (reference a)) = (expression (reference b))))");
		assertParse("a = a<>b", "(field a = (expression (expression (reference a)) <> (expression (reference b))))");
		
		assertParse("a = a or b", "(field a = (expression (expression (reference a)) or (expression (reference b))))");
		assertParse("a = a and b", "(field a = (expression (expression (reference a)) and (expression (reference b))))");
		assertParse("a = a nor b", "(field a = (expression (expression (reference a)) nor (expression (reference b))))");
		assertParse("a = a nand b", "(field a = (expression (expression (reference a)) nand (expression (reference b))))");
		assertParse("a = a xor b", "(field a = (expression (expression (reference a)) xor (expression (reference b))))");
	}
	
	@Test
	public void testLiteralExp() {
		setCurrentRule("field");
		assertParse("a = 12", "(field a = (expression (literal (integerLiteral 12))))");
	}
	
	@Test
	public void testIndexedExp() {
		setCurrentRule("field");
		assertParse("a=f[2]", "(field a = (expression (expression (reference f)) [ (expression (literal (integerLiteral 2))) ]))");
		assertParse("a=b[x][6]");
	}
	
	@Test
	public void testSelectedExp() {
		setCurrentRule("field");
		assertParse("a=x.y", "(field a = (expression (expression (reference x)) . y))");
		assertParse("a=x.y[4]", "(field a = (expression (expression (expression (reference x)) . y) [ (expression (literal (integerLiteral 4))) ]))");
		assertParse("a=x.y.z.t");
	}
	
	@Test
	public void testTransition() {
		setCurrentRule("transition");
		assertParse("from a loop", "(transition from a (block (statement (loopStatement loop))))");
	}
	
	@Test
	public void testInputStmt() {
		setCurrentRule("transition");
		assertParse("from a input x", "(transition from a (block (statement (inputStatement input (portName x)))))");
		assertParse("from b input y(data)", "(transition from b (block (statement (inputStatement input (portName y) ( (expression (reference data)) )))))");
	}
	
	@Test
	public void testOutputStmt() {
		setCurrentRule("transition");
		assertParse("from a output x", "(transition from a (block (statement (outputStatement output (portName x)))))");
		assertParse("from b output y(data)", "(transition from b (block (statement (outputStatement output (portName y) ( (expression (reference data)) )))))");
	}
	
	@Test
	public void testToStmt() {
		setCurrentRule("transition");
		assertParse("from a to b", "(transition from a (block (statement (toStatement to b))))");
		assertParse("from a to xy");
		denyParse("from a to 2+3");
	}
	
	@Test
	public void testLoopStmt() {
		setCurrentRule("transition");
		assertParse("from a loop", "(transition from a (block (statement (loopStatement loop))))");
	}
	
	@Test
	public void testIfStmt() {
		setCurrentRule("transition");
		assertParse("from a if x=2 then to b end if");
		assertParse("from a if x=2 then to b elsif y = 3 then to c end if");
		assertParse("from a if x=2 then to b  elsif y = 3 then to c elsif y = 4 then to c end");
		assertParse("from a if x=2 then to b else loop end if");
		assertParse("from a if x=2 then to b elsif 2=3 then input e else output b end if");
		assertParse("from a if x=2 or y = 7 then to b else if y=3 then loop elsif i then to d end /*inner if*/ end /*outer if*/");
	}
	
	@Test
	public void testSelectStmt() {
		setCurrentRule("transition");
		assertParse("from b select loop [] output b end", "(transition from b (block (statement (selectStatement select (block (statement (loopStatement loop))) [] (block (statement (outputStatement output (portName b)))) end))))");
	}
	
	@Test
	public void testTimerStmt() {
		setCurrentRule("transition");
		assertParse("from b tim.arm(5)");
		assertParse("from c tim.stop");
	}
	
	@Test
	public void testWaitStmt() {
		setCurrentRule("transition");
		assertParse("from x wait [s ,e]", "(transition from x (block (statement (waitStatement wait [ (expression (reference s)) , (expression (reference e)) ]))))");
	}
	
	@Test
	public void testAssignStmt() {
		setCurrentRule("transition");
		assertParse("from a x:=3", "(transition from a (block (statement (assignStatement (expression (reference x)) := (expression (literal (integerLiteral 3)))))))");
	}
	
	@Test
	public void testSequenceStmt() {
		setCurrentRule("transition");
		assertParse("from a loop; input myport");
	}
	
	@Test public void test0() { assertFile(new File("./examples/example1-RW.abcd").getPath()); }
	@Test public void test1() { assertFile(new File("./examples/example10-stmt.abcd").getPath()); }
	@Test public void test2() { assertFile(new File("./examples/example11-par.abcd").getPath()); }
	@Test public void test3() { assertFile(new File("./examples/example12-min.abcd").getPath()); }
	@Test public void test4() { assertFile(new File("./examples/example13-send_receive.abcd").getPath()); }
	@Test public void test5() { assertFile(new File("./examples/example14-counter.abcd").getPath()); }
	@Test public void test6() { assertFile(new File("./examples/example15_multicoms.abcd").getPath()); }
	@Test public void test7() { assertFile(new File("./examples/example16_local_mutex.abcd").getPath()); }
	@Test public void test8() { assertFile(new File("./examples/example16_mutex.abcd").getPath()); }
	@Test public void test9() { assertFile(new File("./examples/example16_nolock_mutex.abcd").getPath()); }
	@Test public void test10() { assertFile(new File("./examples/example16_notused_mutex.abcd").getPath()); }
	@Test public void test11() { assertFile(new File("./examples/example2-commSync.abcd").getPath()); }
	@Test public void test12() { assertFile(new File("./examples/example3-commASync.abcd").getPath()); }
	@Test public void test13() { assertFile(new File("./examples/example3_async_2chan_b.abcd").getPath()); }
	@Test public void test14() { assertFile(new File("./examples/example3_async_2chan_bnb.abcd").getPath()); }
	@Test public void test15() { assertFile(new File("./examples/example3_async_2chan_nb.abcd").getPath()); }
	@Test public void test16() { assertFile(new File("./examples/example3_async_block.abcd").getPath()); }
	@Test public void test17() { assertFile(new File("./examples/example3_async_noblock.abcd").getPath()); }
	@Test public void test18() { assertFile(new File("./examples/example3_async_sync.abcd").getPath()); }
	@Test public void test19() { assertFile(new File("./examples/example3_async_sync_noblock.abcd").getPath()); }
	@Test public void test20() { assertFile(new File("./examples/example4-comput.abcd").getPath()); }
	@Test public void test21() { assertFile(new File("./examples/example5-prodCons.abcd").getPath()); }
	@Test public void test22() { assertFile(new File("./examples/example5_prodCons_nomulti.abcd").getPath()); }
	@Test public void test23() { assertFile(new File("./examples/example6-stack.abcd").getPath()); }
	@Test public void test24() { assertFile(new File("./examples/example7-mutex.abcd").getPath()); }
	@Test public void test25() { assertFile(new File("./examples/example7-semaphore.abcd").getPath()); }
	@Test public void test26() { assertFile(new File("./examples/example7_local_semaphore.abcd").getPath()); }
	@Test public void test27() { assertFile(new File("./examples/example7_notused_semaphore.abcd").getPath()); }
	@Test public void test28() { assertFile(new File("./examples/example8-types.abcd").getPath()); }
	@Test public void test29() { assertFile(new File("./examples/example9-expressions.abcd").getPath()); }
	@Test public void test30() { assertFile(new File("./examples/example_async_chan.abcd").getPath()); }
	@Test public void test31() { assertFile(new File("./examples/example_defer_lowlevel.abcd").getPath()); }
	@Test public void test32() { assertFile(new File("./examples/example_endloop.abcd").getPath()); }
	@Test public void test33() { assertFile(new File("./examples/example_fct1.abcd").getPath()); }
	@Test public void test34() { assertFile(new File("./examples/example_fifo_ops.abcd").getPath()); }
	@Test public void test35() { assertFile(new File("./examples/example_foreach.abcd").getPath()); }
	@Test public void test36() { assertFile(new File("./examples/example_ping_pong.abcd").getPath()); }
	@Test public void test37() { assertFile(new File("./examples/example_portmap.abcd").getPath()); }
	@Test public void test38() { assertFile(new File("./examples/example_sync_chan.abcd").getPath()); }
	@Test public void test39() { assertFile(new File("./examples/example_while.abcd").getPath()); }
	@Test public void test40() { assertFile(new File("./examples/timer_api.abcd").getPath()); }
	@Test public void test41() { assertFile(new File("./examples/timer_api_local.abcd").getPath()); }
	@Test public void test42() { assertFile(new File("./examples/timer_api_shared.abcd").getPath()); }
	@Test public void test43() { assertFile(new File("./examples/timer_flat.abcd").getPath()); }
	@Test public void test44() { assertFile(new File("./examples/timer_shared.abcd").getPath()); }
	
	class ABCDFileFilter implements FilenameFilter {
		public boolean accept(File dir, String name) {
			return name.endsWith(".abcd");
		}
	}
	
	public ParseTree assertParse(String input) {
		ParseTree result = parse(input, currentRule);
		assertTrue(result != null);
		return result;
	}
	
	public ParseTree assertParse(String input, String expectation) {
		ParseTree result = parse(input, currentRule);
		assertTrue(result != null);
		
		Object actual = result.toStringTree(mParser);
		assertEquals("testing " + input + " gives " + expectation, expectation, actual);
		return result;
	}
	
	public ParseTree assertFile(String filename) {
		ParseTree result = parseFile(filename);
		assertTrue("testing "+ filename, result != null);
		return result;
	}
	
	public ParseTree denyParse(String input) {
		ParseTree result = parse(input, currentRule);
		assertTrue(result == null);
		return result;
	}
	
	public ParseTree parse(CharStream cs, String rule) {
		ABCDLexer 			lexer 	= new ABCDLexer(cs);
		CommonTokenStream 	tokens 	= new CommonTokenStream(lexer);
		NoErrorsForTest  	errorL  = new NoErrorsForTest();
		Parser 				parser 	= new ABCDParser(tokens);

		mParser = parser;
		parser.removeErrorListeners();
		parser.addErrorListener(errorL);
		try {
			Method mtd = parser.getClass().getMethod(rule);
			ParseTree pt = (ParseTree)mtd.invoke(parser);
			if (errorL.hasErrors()) return null;
			return pt; 
		} catch (Exception e) {
			System.err.println("no matching method for rule: " + rule);
			return null;
		}
	}

	public ParseTree parse(String input, String rule) {
		ANTLRInputStream 	is		= new ANTLRInputStream(input);
		return parse(is, rule);
	}
	
	public ParseTree parseFile(String filename) {
		ANTLRFileStream fs;
		try {
			fs = new ANTLRFileStream(filename);
		} catch (IOException e) {
			e.printStackTrace();
			return null;
		}
		return parse(fs, "system");
	}
	
	public String getCurrentRule() {
		return currentRule;
	}

	public void setCurrentRule(String currentRule) {
		this.currentRule = currentRule;
	}

	public static class NoErrorsForTest extends BaseErrorListener {
		private Boolean hasErrors = false;
		@Override
		public void syntaxError(Recognizer<?, ?> rec, Object offendingSymbol, int line, int column, String msg, RecognitionException e) {
			hasErrors = true;
		}
		public Boolean hasErrors() {
			return hasErrors;
		}
	}
}

